# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Plugin module for imminent engine failure detection """

from keras.models import load_model
from keras import backend as K
import tensorflow as tf
import pandas as pd
import numpy as np
import os
import joblib
import json
import datetime
import copy
import logging

from foglamp.common import logger

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2019 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Notification rule plugin which detects imminent engine failure',
        'type': 'string',
        'default': 'engine_failure',
        'readonly': 'true'
    },
    'assetName': {
        'description': 'Name of asset',
        'type': 'string',
        'default': 'EngineFailure',
        'displayName': 'Asset name to monitor'
    },
    'modelDirPath': {
        'description': 'Model dir path',
        'type': 'string',
        'default': os.path.join(os.path.dirname(__file__), "model"),
        'displayName': 'Model directory path'
    }
}


_LOGGER = logger.setup(__name__, level=logging.WARNING)
min_max_scaler = None
cols_normalize = None
sequence_cols = None
sequence_length = None
estimator = None
graph = None
failingEngineNumList = list()
failingEngineCycleList = list()
failingEngineListIdx = 0

columns = ['id', 'cycle', 'setting1', 'setting2', 'setting3', 's1', 's2', 's3',
           's4', 's5', 's6', 's7', 's8', 's9', 's10', 's11', 's12', 's13', 's14',
           's15', 's16', 's17', 's18', 's19', 's20', 's21']
test_df = pd.DataFrame(columns=columns)


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.info("plugin_info called")
    return {
        'name': 'engine_failure',
        'version': '1.8.0',
        'mode': 'poll',
        'type': 'notificationRule',
        'interface': '1.0.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config):
    """ Initialise the plugin.
    Args:
        config: JSON configuration document for the plugin configuration category
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """
    global min_max_scaler, cols_normalize, sequence_cols, sequence_length, estimator, graph
    global failingEngineNumList, failingEngineCycleList, failingEngineListIdx

    _LOGGER.info("plugin_init called: config={}".format(config))
    try:
        handle = copy.deepcopy(config)
        handle['modelPath'] = os.path.join(handle['modelDirPath']['value'], "classification_model.h5")
        handle['modelDataFilePath'] = os.path.join(handle['modelDirPath']['value'], "model-data.npy")
        handle['scalerFilePath'] = os.path.join(handle['modelDirPath']['value'], "scaler.save")

        model_path = handle['modelPath']
        model_data_file_path = handle['modelDataFilePath']
        model_data = np.load(model_data_file_path, allow_pickle=True).item()

        # w1 = model_data['w1']
        # w0 = model_data['w0']
        cols_normalize = model_data['cols_normalize']
        sequence_cols = model_data['sequence_cols']
        sequence_length = model_data['sequence_length']

        scaler_file_path = handle['scalerFilePath']
        min_max_scaler = joblib.load(scaler_file_path)

        if os.path.isfile(model_path):
            t1 = datetime.datetime.now()
            estimator = load_model(model_path)
            graph = tf.get_default_graph()
            t2 = datetime.datetime.now()
            delta = t2 - t1
            _LOGGER.info("ML model loaded in {} msec".format(int(delta.total_seconds() * 1000)))
        else:
            _LOGGER.error("Unable to find ML model to load @ {}".format(model_path))
    except Exception as ex:
        _LOGGER.exception('Error in initializing engine_failure plugin {}'.format(str(ex)))
        raise

    failingEngineNumList = list()
    failingEngineCycleList = list()
    failingEngineListIdx = 0
    return handle


def plugin_triggers(handle):
    """ Returns assets to monitor for engine failure data
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        Returns the dict with asset name(s) under which engine failure data is available. This would be used to
        subscribe to such data from storage service
    """
    # d = {'triggers': [{'asset': handle['assetName']['value']}]}
    d = {'triggers': [{'asset': handle['assetName']['value'], 'All': 5}]}
    _LOGGER.info("plugin_triggers(): returning '{}'".format(d))
    return d


def plugin_eval(handle, data):
    """ Evaluates whether engine failure is imminent by running readings' data thru' ML model
    Args:
        handle: handle returned by the plugin initialisation call
        data: dict with the readings, containing engine data
    Returns:
        Returns a bool indicating whether failure is imminent for some engine
    """
    global test_df, min_max_scaler, cols_normalize, sequence_cols, sequence_length, estimator, graph
    global failingEngineNumList, failingEngineCycleList, failingEngineListIdx

    t1 = datetime.datetime.now()
    data = json.loads(str(data))
    data = data[handle['assetName']['value']]
    if isinstance(data['cycle'], (int, float)):
        chunk = pd.DataFrame(data, index=[0])
    else:
        chunk = pd.DataFrame(data)

    _LOGGER.info("Got {} rows to process".format(len(chunk)))

    # MinMax normalization (from 0 to 1)
    chunk['cycle_norm'] = chunk['cycle']
    norm_chunk = pd.DataFrame(min_max_scaler.transform(chunk[cols_normalize]),
                              columns=cols_normalize,
                              index=chunk.index)
    chunk_join_df = chunk[chunk.columns.difference(cols_normalize)].join(norm_chunk)
    chunk = chunk_join_df.reindex(columns=chunk.columns)
    chunk = chunk.reset_index(drop=True)
    test_df = test_df.append(chunk, sort=False)

    # for id in test_df['id'].unique():
    #     _LOGGER.info("{},{}".format(id, len(test_df[test_df['id']==id])))

    # We pick the last sequence for each id in the test data
    seq_array_test_last = [test_df[test_df['id'] == id_][sequence_cols].values[-sequence_length:]
                           for id_ in test_df['id'].unique() if len(test_df[test_df['id'] == id_]) >= sequence_length]

    ids = [[id_, len(test_df[test_df['id'] == id_])]
           for id_ in test_df['id'].unique() if len(test_df[test_df['id'] == id_]) >= sequence_length]

    seq_array_test_last = np.asarray(seq_array_test_last).astype(np.float32)

    rv = False
    if seq_array_test_last.shape[0]:
        # make prediction
        with graph.as_default():
            y_pred_test = estimator.predict_classes(seq_array_test_last)

        t2 = datetime.datetime.now()
        # delta = t2 - t1
        # _LOGGER.info("ids={}, y_pred_test={}, time taken = {}".format(ids, y_pred_test.tolist(),
        #                                                               int(delta.total_seconds() * 1000)))
        for id_, pred in zip(ids, y_pred_test):
            if pred == 1:
                # _LOGGER.info("id={}, pred={}".format(id_, pred))
                if not id_[0] in failingEngineNumList:  # store the engine # that is failing
                    failingEngineNumList.append(id_[0])
                    cycle = test_df[test_df['id'] == id_[0]]['cycle'].max()
                    failingEngineCycleList.append(cycle)
                    # _LOGGER.info("failingEngineNumList={}".format(str(failingEngineNumList)[1:-1]))
                    # _LOGGER.info("failingEngineCycleList={}".format(failingEngineCycleList))
                    rv = True

    # trim rows for previous engine(s)
    ids = test_df['id'].unique()
    for id_ in ids[:-1]:
        test_df = test_df.iloc[len(test_df[test_df['id'] == id_]):]

    return rv


def plugin_reason(handle):
    """ Returns reason string for the last time plugin_eval returned True
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        Returns a simple dict with 'reason' field indicating the reason for last 'plugin_eval' call returning True
    """
    global failingEngineNumList, failingEngineCycleList, failingEngineListIdx

    failedEngines = failingEngineNumList[failingEngineListIdx:]
    failedEnginesCycle = failingEngineCycleList[failingEngineListIdx:]
    s = "triggered for"
    for i in range(0, len(failedEngines)):
        s += " engine {} (cycle {}),".format(failedEngines[i], failedEnginesCycle[i])
    s = s[:-1]
    failingEngineListIdx = len(failingEngineNumList)
    _LOGGER.info("plugin_reason(): failingEngineListIdx={}, reason={}".format(failingEngineListIdx, s))
    return {'reason': s}


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin
    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _LOGGER.info("Old config for engine_failure plugin={} \n new config={}".format(handle, new_config))
    plugin_shutdown(handle)
    new_handle = plugin_init(new_config)
    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    global test_df, min_max_scaler, cols_normalize, sequence_cols, sequence_length, estimator, graph, columns

    _LOGGER.info('engine_failure plugin shut down.')
    K.clear_session()
    min_max_scaler = None
    cols_normalize = None
    sequence_cols = None
    sequence_length = None
    # estimator = None
    # graph = None
    test_df = pd.DataFrame(columns=columns)
